/* Getting started pages */

/* data intro */
.gs-data {
  font-size: 0.9rem;
}

.gs-data-title {
  align-items: center;
  font-size: 0.9rem;
}

.gs-data-title .badge {
  margin: 10px;
  padding: 5px;
}

.gs-data .badge {
  cursor: pointer;
  padding: 10px;
  border: none;
  text-align: left;
  outline: none;
  font-size: 12px;
}

.gs-data .btn {
 background-color: grey;
 border: none;
}

/* note/alert properties */

.alert-heading {
  font-size: 1.2rem;
}

/* callout properties */
.gs-callout {
  padding: 20px;
  margin: 20px 0;
  border: 1px solid #eee;
  border-left-width: 5px;
  border-radius: 3px;
}
.gs-callout h4 {
  margin-top: 0;
  margin-bottom: 5px;
}
.gs-callout p:last-child {
  margin-bottom: 0;
}
.gs-callout code {
  border-radius: 3px;
}
.gs-callout+.gs-callout {
  margin-top: -5px;
}
.gs-callout-remember {
  border-left-color: #f0ad4e;
  align-items: center;
  font-size: 1.2rem;
}
.gs-callout-remember h4 {
  color: #f0ad4e;
}

/* reference to user guide */
.gs-torefguide {
  align-items: center;
  font-size: 0.9rem;
}

.gs-torefguide .badge {
  background-color: #130654;
  margin: 10px 10px 10px 0px;
  padding: 5px;
}

.gs-torefguide a {
  margin-left: 5px;
  color: #130654;
  border-bottom: 1px solid #FFCA00f3;
  box-shadow: 0px -10px 0px #FFCA00f3 inset;
}

.gs-torefguide p {
  margin-top: 1rem;
}

.gs-torefguide a:hover {
  margin-left: 5px;
  color: grey;
  text-decoration: none;
  border-bottom: 1px solid #b2ff80f3;
  box-shadow: 0px -10px 0px #b2ff80f3 inset;
}

/* question-task environment */

ul.task-bullet, ol.custom-bullet{
  list-style:none;
  padding-left: 0;
  margin-top: 2em;
}

ul.task-bullet > li:before {
    content:"";
    height:2em;
    width:2em;
    display:block;
    float:left;
    margin-left:-2em;
    background-position:center;
    background-repeat:no-repeat;
    background-color: #130654;
    border-radius: 50%;
    background-size:100%;
    background-image:url('../question_mark_noback.svg');
  }

ul.task-bullet > li {
  border-left: 1px solid #130654;
  padding-left:1em;
}

ul.task-bullet > li > p:first-child {
  font-size: 1.1rem;
  padding-left: 0.75rem;
}

/* Getting started index page */

.comparison-card {
  background:#FFF;
  border-radius:0;
  padding: 30px 10px 10px 10px;
  margin: 10px 0px;
}

.comparison-card p.card-text {
  margin: 0px;
}

.comparison-card .card-img-top {
  margin: 10px;
  margin-bottom: 20px;
  height: 72px;
}

.comparison-card-excel .card-img-top, .comparison-card-stata .card-img-top, .comparison-card-sas .card-img-top {
  height: 52px;
}

.comparison-card .card-footer {
  border: none;
  background-color:white;
}

.install-block {
  padding-bottom: 30px;
}

.install-card .card-header {
  border: none;
  background-color:white;
  padding: 1rem 1rem 0rem 1rem;
}

.install-card .card-header p.card-text {
  color: #150458;
  font-size: 1.1rem;
  font-weight: bold;
}

.install-card .card-footer {
  border: none;
  background-color:white;
}

.install-card pre {
  margin: 0 1em 1em 1em;
}

.custom-button {
  background-color:#DCDCDC;
  border: none;
  color: #484848;
  text-align: center;
  text-decoration: none;
  display: inline-block;
  font-size: 0.9rem;
  border-radius: 0.5rem;
  max-width: 120px;
  padding: 0.5rem 0rem;
}

.custom-button a {
  color: #484848;
}

.custom-button p {
  margin-top: 0;
  margin-bottom: 0rem;
  color: #484848;
}

/* intro to tutorial collapsed cards */

.tutorial-accordion {
  margin-top: 20px;
  margin-bottom: 20px;
}

.tutorial-card .card-header.card-link .btn {
  margin-right: 12px;
}

.tutorial-card .card-header.card-link .btn:after {
  content: "-";
}

.tutorial-card .card-header.card-link.collapsed .btn:after {
  content: "+";
}

.tutorial-card-header-1 {
  justify-content: space-between;
  align-items: center;
}

.tutorial-card-header-2 {
  justify-content: flex-start;
  align-items: center;
  font-size: 1.3rem;
}

.tutorial-card .card-header {
  cursor: pointer;
  background-color: white;
}

.tutorial-card .card-body {
  background-color: #F0F0F0;
}

.tutorial-card .badge {
  background-color: #130654;
  margin: 10px 10px 10px 10px;
  padding: 5px;
}

.tutorial-card .gs-badge-link p {
  margin: 0px;
}

.tutorial-card .gs-badge-link a {
  color: white;
  text-decoration: none;
}

.tutorial-card .badge:hover {
  background-color: grey;
}
